#include "tivoDeclIter.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "tivoCallGraph.h"
#include "cp/cp-tree.h"
#include "tivoTable.h"

#define CG_PREFIX "CALL_GRAPH: "

static FILE* pCgOut = NULL;
static int   fMustClose = 0;

static int DemangleNames()
{
    static int fSet = 0;
    static int fDemangle = 0;

    if ( !fSet )
    {
        fDemangle = ( getenv("TIVO_CALL_GRAPH_DEMANGLE") != NULL );
        fSet = 1;
    }

    return fDemangle;
}

FILE* GetCgOut()
{
    static int   fSet = 0;
    static char* path = NULL;
    int fAppend = 0;

    if ( fSet )
        return pCgOut;

    // we only want to compute this once...
    fSet = 1;

    path = getenv( "TIVO_CALL_GRAPH" );
    fAppend = (getenv( "TIVO_CALL_GRAPH_APPEND" ) != NULL);

    if ( NULL == path )
        return pCgOut;

    if ( strcmp( path, "stdout" ) == 0 )
    {
        pCgOut = stdout;
    }
    else if ( strcmp( path, "stderr" ) == 0 )
    {
        pCgOut = stderr;
    }
    else
    {
        pCgOut = fopen( path, fAppend ? "a" : "w" );
        if ( NULL == pCgOut )
        {
            fprintf( stderr,
                     CG_PREFIX "can't open '%s' for call graph!\n",
                     path );
        }
        else
        {
            fMustClose = 1;
        }
    }
    //fprintf( stderr, "CG path = %s (%p)\n", path, pCgOut );

    return pCgOut;
}

void tivoCallGraphClose()
{
    //fprintf( stderr, "tivoCallGraphClose()  %d %p\n", fMustClose, pCgOut );
    if ( fMustClose && pCgOut != NULL )
    {
        fclose( pCgOut );
    }

    pCgOut = NULL;
}

static const char* GetHumanName( tree node )
{
    return (*decl_printable_name)( node, 2 );
}

static const char* GetMangledName( tree node )
{
    tree mangledNode = DECL_ASSEMBLER_NAME( node );
    const char* mangled = mangledNode ? IDENTIFIER_POINTER( mangledNode ) : NULL;

    return mangled ? mangled : "NULL";
}

static const char* GetName( tree node )
{
    if ( TREE_CODE( node ) == RECORD_TYPE || DemangleNames() )
    {
        return GetHumanName( node );
    }
    else
    {
        return GetMangledName( node );
    }
}

// get the table of code names!
extern char **tree_code_name;

int GetVtableIndex( tree intConst )
{
    int vtblIndex = TREE_INT_CST_LOW(intConst);
    if (TREE_INT_CST_HIGH(intConst) != 0)
    {
        if ( GetCgOut() )
            fprintf( GetCgOut(), CG_PREFIX "vtable index seems too big %d %d!\n",
                     TREE_INT_CST_HIGH(intConst),
                     TREE_INT_CST_LOW(intConst) );
    }
    return vtblIndex;
}

const char* tivoGetId( tree node )
{
    char class = TREE_CODE_CLASS (TREE_CODE (node));

    if (class == 'd')
    {
        if ( DECL_NAME(node) )
        {
            return IDENTIFIER_POINTER( DECL_NAME(node) );
        }
    }
    else if (class == 't')
    {
        if ( TYPE_NAME(node) )
	{
            if ( TREE_CODE( TYPE_NAME(node) ) == IDENTIFIER_NODE )
            {
                return IDENTIFIER_POINTER( TYPE_NAME(node) );
            }
            else if (TREE_CODE (TYPE_NAME (node)) == TYPE_DECL
                     && DECL_NAME (TYPE_NAME (node)))
            {
                return IDENTIFIER_POINTER( DECL_NAME( TYPE_NAME( node ) ) );
            }
	}
    }

    if ( TREE_CODE(node) == IDENTIFIER_NODE )
    {
        return IDENTIFIER_POINTER(node);
    }

    return "*nullName*";
}

const char* tivoGetFunctionName( tree function_decl )
{
    const char* name = "*topLevel*";

    if ( NULL != function_decl )
    {
        /* Allow functions to be nameless (such as artificial ones).  */
        //fprintf( stderr, ">>>>> %p\n", function_decl );
        if ( TREE_CODE(function_decl) == FUNCTION_DECL )
        {
            name = GetName(function_decl);
        }
        //fprintf( stderr, "<<<<<\n" );
    }

    return name;    
}

/**
 * if expr is NULL, returns NULL
 * if expr isn't the given treeCode, prints an error, and returns NULL
 * otherwise, returns the ith argument.
 */
static tree GetArg( tree expr, enum tree_code treeCode, int i )
{
    enum tree_code gotCode = 0;
    if ( NULL == expr )
        return NULL_TREE;

    gotCode = TREE_CODE( expr );
    if ( gotCode != treeCode )
    {
        if ( GetCgOut() )
            fprintf( GetCgOut(), CG_PREFIX "expected type %s (%d), got %s (%d)\n",
                     tree_code_name[treeCode], treeCode,
                     tree_code_name[gotCode], gotCode );
        return NULL_TREE;
    }

    return TREE_OPERAND( expr, i );
}

void tivoReportFunctionCall( tree callee_function )
{
    const char* caller = "";
    const char* callee = "";
    tree function = NULL;

    if ( ! GetCgOut() )
    {
        return;
    }

    caller = tivoGetFunctionName( current_function_decl );

    // sometimes, i pass in an addr_expr containing a function_decl
    // if i've done that, grab the function_decl!

    function = callee_function;

    //debug_tree( function );
    //printf( "\n\n\n\n\n\n\n\n" );

    if ( TREE_CODE (function) == INDIRECT_REF )
    {
        tree plusExpr    = GetArg( function,    INDIRECT_REF, 0 );
        tree nopExpr     = GetArg( plusExpr,    PLUS_EXPR, 0 );
        tree compRef     = GetArg( nopExpr,     NOP_EXPR, 0 );
        tree indirectRef = GetArg( compRef,     COMPONENT_REF, 0 );
        tree intConst    = GetArg( plusExpr,    PLUS_EXPR, 1 );

        if ( indirectRef && intConst )
        {
            tree type = TREE_TYPE( indirectRef );
            int vtblIndex = GetVtableIndex( intConst );
            if ( GetCgOut() )
                fprintf( GetCgOut(), "vcall {%s} --> {%s} %d\n",
                         caller,
                         tivoGetId( type ), vtblIndex/4 );
        }
        else
        {
            if ( GetCgOut() )
                fprintf( GetCgOut(), CG_PREFIX "internal error!" );
        }

        return;
    }

    if ( TREE_CODE (function) == ADDR_EXPR &&
         TREE_CODE (TREE_OPERAND (function, 0)) == FUNCTION_DECL )
    {
        function = TREE_OPERAND (function, 0);
    }

    if ( function )
    {
        callee = tivoGetFunctionName( function );

        if ( GetCgOut() )
            fprintf( GetCgOut(), "call {%s} --> {%s}\n", caller, callee );
    }
}
 
/******************************
 *
 * Check the global scope...
 *
 */

static int nIndent = 0;

#define INDENT_DIFF 3

static const char* Indent( int nIndent )
{
#define BUFLEN 300
    static char buf[BUFLEN];
    static int  curIndent = -1;

    if ( curIndent != nIndent )
    {
        int i = 0;
        for( i=0; (i < nIndent) && (i < (BUFLEN-1)); i++ )
        {
            buf[i] = ' ';
        }
        buf[i] = '\0';
        curIndent = nIndent;
    }

    return buf;
#undef BUFLEN
}

int tivoGenClassGraph( tree node, int fStart )
{
    // get the node's code to dispatch below!
    int nodeCode = TREE_CODE( node );

    assert( GetCgOut() != NULL && "should NOT have called!" );

    /*
    fprintf( GetCgOut(), "---> tivoGenClassGraph: %s(%s)\n",
	     tree_code_name[nodeCode],
	     tivoGetId(node) );
    */

    if ( nodeCode == VAR_DECL )
    {
    }
    else if ( nodeCode == FIELD_DECL )
    {
    }
    else if ( nodeCode == FUNCTION_DECL )
    {
        tree vindex = DECL_VINDEX( node );
        int vtblIndex = -1;
        int fVirtual = 0;
        tree context = DECL_CONTEXT( node );
        const char* className = NULL;
        
        if ( vindex )
        {
            if ( TREE_CODE( vindex ) == INTEGER_CST )
            {
                vtblIndex = GetVtableIndex( vindex );
                fVirtual = 1;
            }
        }
        
        if ( NULL != context )
            className = GetName(context);
        
        fprintf( GetCgOut(), "%s", Indent( nIndent ) );
        if ( fVirtual )
            fprintf( GetCgOut(), "v" );
        
        fprintf( GetCgOut(), "function " );
        
        if ( fVirtual )
            fprintf( GetCgOut(), "%d ", vtblIndex );
        
        if ( NULL != className )
            fprintf( GetCgOut(), "in {%s} ", className );
        
        fprintf( GetCgOut(), "{%s}\n", tivoGetFunctionName(node) );
    }
    else if ( nodeCode == TYPE_DECL )
    {
    }
    else if ( nodeCode == RECORD_TYPE )
    {
        if ( fStart )
        {
            if ( NULL != node &&
                 tivoIsClassType(node) &&
                 TYPE_BINFO(node) != NULL)
            {
                fprintf( GetCgOut(), "record {%s}:",
                         GetName(node));
                {
                    int nBases = CLASSTYPE_N_BASECLASSES(node);
                    int iBase  = 0;
                    tree binfos = TYPE_BINFO_BASETYPES(node);
                    
                    for( iBase=0; iBase < nBases; iBase++ )
                    {
                        tree aBase = BINFO_TYPE( TREE_VEC_ELT( binfos, iBase ) );
                        fprintf( GetCgOut(), "{%s} ",
                                 GetName(aBase) );
                    } 
                }
                fprintf( GetCgOut(), "\n" );
            }
        }
    }
    else if ( nodeCode == TREE_VEC )
    {
    }
    else
    {
    }

    return 1;  /* we always recurse into subtrees! */
}

void tivoGenCallGraphForGlobal()
{
    if ( NULL != GetCgOut() ) {
        tivoIterateOverGlobal( tivoGenClassGraph );
    }
}


void tivoReportAddressOf( tree arg )
{
    if ( GetCgOut() == NULL )
        return;

    if ( FUNCTION_DECL == TREE_CODE( arg ) )
    {
        const char* caller = tivoGetFunctionName( current_function_decl );
        const char* callee = GetName( arg );

        fprintf( GetCgOut(), "addressOf in {%s} --> {%s}\n", caller, callee );
        fprintf( GetCgOut(), "addressOf in {%s} --> {%s}\n", caller, callee );
    }
}

